#include "bgCore.h"
#include "Platform.h"
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

// accessor
uint8_t NodeDepth(BAT node) {
  Node *n = LoadBlock(node);
  int depth = n->depth;
  UnloadBlock(node, n);
  return depth;
}

// accessor
int NodeIsLeaf(BAT node) {
  return NodeDepth(node) == 0;
}

// accessor
BAT NodeItem(BAT node, int i) {
  Node *n = LoadBlock(node);
  BAT item = n->items[i];
  UnloadBlock(node, n);
  return item;
}

// accessor
BAT NodeKey(BAT node, int i) {
  Node *n = LoadBlock(node);
  BAT key = n->keys[i];
  UnloadBlock(node, n);
  return key;
}

// accessor
BAT NodeParent(BAT node) {
  Node *n = LoadBlock(node);
  BAT parent = n->parentNode;
  UnloadBlock(node, n);
  return parent;
}

/* Inefficient, should probably stash parent index in Node
 * structure. */
// accesses on parent: NodeLen, items[]
uint16_t FindIndexOfChild(BAT parent, BAT child) {
  Node *n = LoadBlock(parent);
  int l = GetNodeLen(parent);
  for (int i=0; i<l; ++i) {
    if (n->items[i] == child) {
      UnloadBlock(parent, n);
      return i;
    }
  }
  UnloadBlock(parent, n);
  return (uint16_t) -1;
}

// accesses on nodeAddr: len, depth, keys, item[i] (if match)
BAT FindInternal(BAT nodeAddr, KeyT key, BAT *outParent, int *outIndex) {
  //fprintf(stderr, "Find %08x ",(unsigned int) nodeAddr);
  //StackPrintInfo(stderr);
  Node *node = LoadBlock(nodeAddr);
  int n = GetNodeLen(nodeAddr);
  if (node->depth == 0) { // leaf nodes special case
    for (int i=0; (i<n) && (key >= node->keys[i]); i++) {
      if (key == node->keys[i]) { // found it!
        *outParent = nodeAddr;
        *outIndex = i;
        BAT value = node->items[i];
        UnloadBlock(nodeAddr, node);
        return value;
      }
    }
    // didn't find it. fail.
    UnloadBlock(nodeAddr,node);
    return BAInvalid;
  }

  int i; // position us so that i is the child we want
  for (i = 0; (i < n) && (key > node->keys[i]); i++)
    ;
  if (i >= n)
    return BAInvalid;
  BAT child = node->items[i]; // -1?
  UnloadBlock(nodeAddr, node);
  return FindInternal(child, key, outParent, outIndex);
}

void InitNode(Node *n) {
  int i;
  n->parentNode = BAInvalid;
  n->depth = 0;
  for (i=0; i<ItemsPerNode; ++i)
    n->items[i] = BAInvalid;
}

/** Get last item in a node **/
// accesses len, items[len-1]
BAT NodeLastItem(BAT node) {
  int len = GetNodeLen(node);
  if (len == 0)
    return BAInvalid;
  Node *nnode = LoadBlock(node);
  BAT last = nnode->items[len - 1];
  UnloadBlock(node, nnode);
  return last;
}

/** Returns the last leaf node under the spcified node.  depth = 0 = leaf node **/
// accesses on node: depth
BAT GetLastNodeOfDepth(BAT node, uint8_t depth) {
  if (node == BAInvalid)
    return BAInvalid;
  if (NodeDepth(node) == depth)
    return node;
  else
    return GetLastNodeOfDepth(NodeLastItem(node), depth);
}


/** Returns last non-full non-leaf node, or invalid if there is none **/
// accesses on node: IsLeaf, len
BAT GetLastNonFullDescendant(BAT node) {
  if (node == BAInvalid)
    return BAInvalid;
  if (NodeIsLeaf(node))
    return BAInvalid;
  BAT nonFull = GetLastNonFullDescendant(NodeLastItem(node));
  if ((nonFull == BAInvalid)
      && (GetNodeLen(node) < ItemsPerNode)) {
    return node;
  }
  return nonFull;
}

/** Returns number of valid internal elements in a node **/
// accesses items (inefficient enumeration)
int GetNodeLen(BAT nodeAddr) {
  Node *node = LoadBlock(nodeAddr);
  for (int i=0; i<ItemsPerNode; i++)
    if (node->items[i] == BAInvalid) { // FIXME: should this be keys[i]?
      UnloadBlock(nodeAddr, node);
      return i;
    }
  UnloadBlock(nodeAddr, node);
  return ItemsPerNode;
}

// accesses depth, InitNode(), SaveBlock, keys, items, everythgin
BAT CreateDescendingNodes(int depth, KeyT lowest) {
  BAT n = AllocBlock();
  Node *nn = LoadBlock(n);
  InitNode(nn);
  nn->depth = depth;
  SaveBlock(n, nn);
  if (depth == 0) {
    UnloadBlock(n, nn);
    return n;
  }
  BAT child = CreateDescendingNodes(depth - 1, lowest);
  nn->keys[0] = lowest;
  nn->items[0] = child;
  UnloadBlock(n, nn);
  Node *nchild = LoadBlock(child);
  nchild->parentNode = n;
  UnloadBlock(child, nchild);
  return n;
}

/** Adds a new node to a full tree **/
BAT AppendNewLeafNode(BAT topNode) {
  BAT nonFull = GetLastNonFullDescendant(topNode);
  if (nonFull == BAInvalid) {
    return BAInvalid; /** The tree is entirely full.  You'd need to add levels.  We dont support that. **/
  }
  int nfDepth = NodeDepth(nonFull);
  /* Build out string of nodes to the leaf level */
  BAT newNodes = CreateDescendingNodes(nfDepth - 1, 0);
  {
    Node *nfNode = LoadBlock(nonFull);
    nfNode->keys[GetNodeLen(nonFull)] = 0; // FIXME
    nfNode->items[GetNodeLen(nonFull)] = newNodes;
    UnloadBlock(nonFull, nfNode);
    Node *nNewNodes = LoadBlock(newNodes);
    nNewNodes->parentNode = nonFull;
    UnloadBlock(newNodes, nNewNodes);
  }
  /* Return the leaf level node of the new nodes */
  BAT n = newNodes;
  for (int i=nfDepth - 2; i >= 0; i--)
    n = NodeItem(n, 0);
  return n;
}

/** A traversal which invokes NodeVisitor on each node.  If NodeVisitor returns
 * a non-zero value, then the traversal is terminated and the current node is
 * returned to the caller **/
BAT NodeVisitDFS(NodeVisitor func, BAT node) {
  if (node == BAInvalid)
    return BAInvalid;
  /** Traverse children **/
  int n = GetNodeLen(node);
  for (int i=0; i<n; i++) {
    if (NodeDepth(node) != 0)
      NodeVisitDFS(func, NodeItem(node, i));    
  }
  /** Call user visitor **/
  {
    Node *nnode = LoadBlock(node);
    int ret = (*func)(node, nnode);
    UnloadBlock(node, nnode);
    if (ret) // Visitor is done, return node where he finished
      return node;
  }
  return BAInvalid;
}

/** Calculates the number of bytes to hold a tree for a set number of elements
 *  May be slightly larger than actually required (by a few blocks) **/
int CalcTreeSize2(int numItems, int depth) {
  int summ=0;
  int n = numItems;
  for (int i=0; i<depth; i++) {
    printf("L%i = %lu\n", i, (unsigned long int) n/ItemsPerNode + 1);
    n = n/ItemsPerNode + 1; /* 1 is a slop factor - could be more precise */
    summ += n;
  }
  return summ * BlockSize;
}

/** Calculate the Minimum tree depth given the number of items to store and the
 * tree parameters.  Might be a little conservative (aka, sloppy). **/
int CalcMinimumDepth(int numItems) {
  int priorn = -999;
  int n = numItems;
  for (int i=0; i<99; i++) {
    n = n/ItemsPerNode + 1; /* 1 is a slop factor - could be more precise */
    if ((n == priorn) && (n == 1))
      return i + 1 + (i == 0);
    priorn = n;
  }
  assert(0);
}

BAT NodeNextLeaf(BAT node, BAT child) {
  if (node == BAInvalid)
    return BAInvalid;
  int index = FindIndexOfChild(node, child);
  assert(index != (uint16_t) -1);
  if (index < GetNodeLen(node)-1) {
    // descend to the left
    BAT nextChild = NodeItem(node, index + 1);
    while (!NodeIsLeaf(nextChild)) {
      nextChild = NodeItem(nextChild, 0);
      assert(nextChild != BAInvalid);
    }
    return nextChild;
  }
  return NodeNextLeaf(NodeParent(node), node);
}

RowPtr RowPtrInit(BgTree table) {
  RowPtr rp;
  rp.tree = table;
  rp.node = BAInvalid;
  rp.index = (uint16_t) -1;
  return rp;
}

RowPtr RowPtrNext(RowPtr row) {
  if (row.node == BAInvalid) {
    return row;
  } if (row.index < GetNodeLen(row.node) - 1) {
    row.index += 1;
    return row;
  }
  BAT nextLeaf = NodeNextLeaf(NodeParent(row.node), row.node);
  row.node = nextLeaf;
  row.index = 0;
  return row;
}

/** Finds the left-most leaf child node (assumes that there IS one) **/
BAT NodeFindFirstChild(BAT node) {
  if (node == BAInvalid)
    return BAInvalid;
  if (NodeDepth(node) == 0)
    return node;
  return NodeFindFirstChild(NodeItem(node, 0));
}
